from typing import Dict, List, Union, cast

from .utils.string_utils import simplify_expression
from .parse_tree import ParseTree
from .utils.mode import Mode
from .utils.io_utils import clear_console, variable_speed_print
from sys import argv


class InteractivePlayground:
    def __init__(self, run_on_init: bool = False, *args, **kwargs) -> None:
        self.__options = args
        self.__current_mode: Union[Mode, List[Mode]] = [
            Mode.DISPLAY, Mode.EVAL]
        self.__mode_dict: Dict[str, Mode] = {
            'a': Mode.ANIMATE,
            'd': Mode.DISPLAY,
            'e': Mode.EVAL
        }
        if run_on_init:
            self.run()

    # Introductory Menu
    def introduction(self) -> None:
        greeting = 'Greetings...\n'\
                   '\n'\
                   '                \"This is a utility program which aims\n'\
                   '                    to show the beauty of parse trees\n'\
                   '                           in a fun and engaging way\"\n'\
                   '\n'\
                   'Don\'t be intimidated :)\n'\
                   'It was designed to be easy to use, yet extensible.\n'\
                   '\n'\
                   '                                               Enjoy!'
        variable_speed_print(message=greeting, seconds_per_char=0.05)

    # Help Menu
    def help(self) -> None:
        help_menu = 'HELP MENU\n'\
                    '=========\n'\
                    'Enter:\n'\
                    '?> help\n'\
                    '         to display this help menu\n'\
                    '?> mode\n'\
                    '         to select your mode'
        print(help_menu)

    def mode(self, ) -> None:
        mode_list = 'CURRENTLY SUPPORTED MODES\n'\
                    '-------------------------\n'\
                    'a. ANIMATE\n'\
                    'd. DISPLAY\n'\
                    'e. EVAL\n\n'\
                    f'current mode: {self.__current_mode}'
        print(mode_list)
        new_mode = simplify_expression(
            input('Enter your desired mode (or permutation thereof): '))
        if new_mode != '':
            if ',' in new_mode:
                self.__current_mode = [self.__mode_dict[m]
                                       for m in new_mode.split(sep=',')]
            else:
                self.__current_mode = self.__mode_dict[new_mode]

    def perform_action(self, parse_tree: ParseTree, action: Mode):
        if action == Mode.ANIMATE:
            parse_tree.animate(1)
        elif action == Mode.DISPLAY:
            print(str(parse_tree))
        elif action == Mode.EVAL:
            print(parse_tree.expression, '=', round(
                cast(Union[int, float], parse_tree.evaluate()), 5))
        print()

    # REPL Simulation

    def run(self):
        '''
        Runs the animated_parse_tree program interactively

        Parameters
        ----------
        **kwargs
            Keyword arguments to be passed to ParseTree().animate.

        Returns
        -------
        None
            This program does not return anything.

        Examples
        --------
        ```bash
        ?> 1 + 2
        1 + 2 = 3

         +
        / \\
        1 2
        ```
        '''
        clear_console()
        if '-q' not in self.__options:
            self.introduction()
        expression = input('?> ')
        while expression not in {'', 'exit', 'quit'}:
            try:
                if expression == 'mode':
                    self.mode()
                elif expression == 'help':
                    self.help()
                else:
                    t = ParseTree()
                    t.read(expression)
                    if type(self.__current_mode) is list:
                        for action in self.__current_mode:
                            self.perform_action(parse_tree=t, action=action)
                    else:
                        self.perform_action(
                            parse_tree=t, action=cast(Mode, self.__current_mode))
            except KeyError as e:
                print(f'Invalid mode {e}')
            except Exception as e:
                print(e)
            finally:
                expression = input('?> ')
        print('Bye :)')


if __name__ == '__main__':
    MAIN = InteractivePlayground(False, *argv)
    MAIN.run()



from copy import deepcopy
from typing import Dict, List, Literal, cast

from .exceptions import LexingError
from .operand_ import Operand
from .operator_ import Operator


class Lexer:
    def __init__(self,
                 token_lookup: Dict,
                 implicit_operator: Operator) -> None:
        self.token_lookup = token_lookup
        self.implicit_operator = implicit_operator

    def lex_token(self, token=None):
        if token is not None:
            if token in self.token_lookup:
                if isinstance(self.token_lookup[token], List):
                    return cast(List[Operator], deepcopy(self.token_lookup[token]))
                else:
                    return cast(Operand, deepcopy(self.token_lookup[token]))
            raise LexingError(f'Unknown token {token} encountered')
        else:
            return deepcopy(self.implicit_operator)

    def __lazy_lexing(self, tokens):
        lexed_token_list = []

        digit_set = set('0123456789.')

        open_parentheses_count = 0
        close_parentheses_count = 0

        for token in tokens:
            if token in '(,)':
                if token == '(':
                    open_parentheses_count += 1
                    lexed_token_list.append(token)
                elif token == ')':
                    close_parentheses_count += 1
                    lexed_token_list.append(token)
                else:
                    lexed_token_list.append(']')
                    lexed_token_list.append('[')
            elif token[0] in digit_set:
                # Normal numeric operand
                num = float(token) if '.' in token else int(token)
                lexed_token_list.append(Operand(value=num))
            else:
                token_obj = self.lex_token(token=token)
                if isinstance(token_obj, Operand):
                    lexed_token_list.append(self.lex_token(token=token))
                elif isinstance(token_obj, List):
                    if len(token_obj) == 1:
                        lexed_token_list.append(
                            cast(List[Operator], self.lex_token(token=token))[0])
                    else:
                        lexed_token_list.append(self.lex_token(token=token))
            yield open_parentheses_count, close_parentheses_count, lexed_token_list, token
        if close_parentheses_count != open_parentheses_count:
            raise LexingError(
                f'{open_parentheses_count} \'(\' does not match {close_parentheses_count} \')\'')

    def __eager_lexing(self, tokens: List[str]) -> List:
        token_obj_list = []
        for _, _, token_objs, _ in self.__lazy_lexing(tokens=tokens):
            token_obj_list = token_objs
        return token_obj_list

    def lex(self, tokens: List[str], kind: Literal['lazy', 'eager'] = 'eager'):
        if kind == 'lazy':
            return self.__lazy_lexing(tokens=tokens)
        elif kind == 'eager':
            return self.__eager_lexing(tokens=tokens)
        else:
            raise ValueError(f'Unknown kind \'{kind}\' received')


from .node import Node


class Temp_Node(Node):
    def __init__(self,
                 parent: Node,
                 blank: str = ' '):
        super().__init__(value=None)
        self.parent = parent
        self.width = parent.width
        self.left_pad = parent.left_pad
        self.right_pad = parent.right_pad
        self.blank: str = blank
        parent.children.append(self)

    def __str__(self):
        return self.blank * (self.left_pad + self.width + self.right_pad)

    def display(self):
        return self.__str__()

    def __repr__(self) -> str:
        return f'TempNode(width={self.width})'


from typing import List, Union
from .operand_ import Operand
from .operator_ import Operator


Bundle = List[Union[Operand, Operator]]


# Basics bundle
BASICS: Bundle = [
    Operator(symbol='+', func=lambda a, b: a + b, priority=1),
    Operator(symbol='-', func=lambda a, b: a - b, priority=1),
    Operator(symbol='*', func=lambda a, b: a * b, priority=3),
    Operator(symbol='/', func=lambda a, b: a / b, priority=3),
    Operator(symbol='-', func=lambda a: -a, priority=9, kind='pre', operands=1)
]


# Constants bundle
from math import pi, e, tau
CONSTANTS: Bundle = [
    Operand(value=pi, symbol='pi'),
    Operand(value=e, symbol='e'),
    Operand(value=tau, symbol='tau')
]


# Exponentiation bundle
from math import log10, log
EXPONENTIATION: Bundle = [
    Operator(symbol='^', func=lambda a, b: a ** b, priority=5),
    Operator(symbol='**', func=lambda a, b: a ** b, priority=5),
    Operator(symbol='log', func=lambda a, b: log(b, a), priority=7, kind='pre', operands=2),
    Operator(symbol='lg', func=lambda a: log10(a), priority=7, kind='pre', operands=1),
    Operator(symbol='ln', func=lambda a: log(a), priority=7, kind='pre', operands=1)
]


# Trigonometry bundle
from math import sin, cos, tan
from math import asin, acos, atan
from math import sinh, cosh, tanh
from math import asinh, acosh, atanh
TRIGONOMETRY: Bundle = [
    Operator(symbol='deg', func=lambda a: pi / 180 * a, priority=7, kind='post', operands=1),
    Operator(symbol='sin', func=lambda a: sin(a), priority=7, kind='pre', operands=1),
    Operator(symbol='sine', func=lambda a: sin( a), priority=7, kind='pre', operands=1),
    Operator(symbol='cos', func=lambda a: cos(a), priority=7, kind='pre', operands=1),
    Operator(symbol='cosine', func=lambda a: cos(a), priority=7, kind='pre', operands=1),
    Operator(symbol='tan', func=lambda a: tan(a), priority=7, kind='pre', operands=1),
    Operator(symbol='tangent', func=lambda a: tan(a), priority=7, kind='pre', operands=1),
]


# Factorial bundle
from math import factorial
FACTORIAL: Bundle = [
    Operator(symbol='!', func=lambda a: factorial(a), priority=9, kind='post', operands=1)
]


# Series bundle
# SERIES: Bundle = [
#     Operator(symbol='sum', func=lambda a, b, c: a + b + c, priority=9)
# ]



from typing import List


def strip_whitespace(string: str) -> str:
    return string.replace(' ', '')


def simplify_expression(expr: str) -> str:
    return strip_whitespace(expr).lower()


def concatenate_horizontally(*args: str) -> str:
    string_list: List[List[str]] = []
    for string in args:
        string_list.append(string.splitlines())
    result = ''
    total_height = max(map(lambda ls: len(ls), string_list))
    for i in range(total_height):
        for ls in string_list:
            try:
                result += ls[i]
            except IndexError:
                pass
        result += '\n'
    return result


def pad(left_pad: int, body: str, right_pad: int, pad_char: str = ' '):
    return pad_char * left_pad + body + pad_char * right_pad



from abc import abstractmethod
from typing import Any, List, Optional
from math import ceil

from .utils.string_utils import pad


class Node:
    def __init__(self,
                 value=None):
        self.parent: Optional[Node] = None
        self.children: List[Node] = []
        self.value: Any = value
        self.height: int = 1
        self.priority: int = -1
        self.inner_width: int = 0
        self.sub_width: int = 0
        self.width: int = 0
        self.bal_coef: int = 0
        self.left_pad: int = 0
        self.right_pad: int = 0

    @abstractmethod
    def isFull(self) -> bool:
        pass

    @abstractmethod
    def isEmpty(self) -> bool:
        pass

    def __lt__(self, otherNode) -> bool:
        return self.priority < otherNode.priority

    def __le__(self, otherNode) -> bool:
        return self.priority <= otherNode.priority

    def __gt__(self, otherNode) -> bool:
        return self.priority > otherNode.priority

    def __ge__(self, otherNode) -> bool:
        return self.priority >= otherNode.priority

    def __call__(self) -> Any:
        return self.value

    @abstractmethod
    def __str__(self) -> str:
        pass

    def display(self) -> str:
        return str(self.value)

    def get_balance_coefficient(self) -> int:
        last_child = self.children[-1]
        right_bal_coef = last_child.bal_coef
        right_space = ceil((last_child.width - abs(right_bal_coef) - 1) / 2) + max(right_bal_coef, 0)

        first_child = self.children[0]
        left_bal_coef = first_child.bal_coef
        left_space = (first_child.width - abs(left_bal_coef) - 1) // 2 - min(left_bal_coef, 0)

        return right_space - left_space


    def branch(self, branch_symbol: str) -> str:
        def __branch():
            if len(self.children) == 0:
                return f'{branch_symbol:^{self.width}}'
            if len(branch_symbol) <= (self.width - abs(self.bal_coef)):
                if self.bal_coef > 0:
                    return f'{branch_symbol:^{self.width - abs(self.bal_coef)}}' + ' ' * abs(self.bal_coef)
                else:
                    return ' ' * abs(self.bal_coef) + f'{branch_symbol:^{self.width - abs(self.bal_coef)}}'
            elif self.bal_coef > 0:
                return f'{branch_symbol:<{self.width}}'
            elif self.bal_coef < 0:
                return f'{branch_symbol:>{self.width}}'
            return f'{branch_symbol:^{self.width}}'
        return pad(left_pad=self.left_pad, body=__branch(), right_pad=self.right_pad)


from typing import Any, Callable, List


def find_element(ls: List, find: Any, default_value: Any, *args):
    try:
        return ls.index(find)
    except ValueError:
        return default_value


def find_element_where(ls: List, condition: Callable[[Any], bool]):
    for el in ls:
        if condition(el):
            return el
    return None


def index_first_element(ls: List, *args):
    return min([find_element(ls=ls, find=el, default_value=len(ls)) for el in args])



from string import ascii_lowercase
from typing import Dict, List, Literal
from .utils.string_utils import simplify_expression
from .exceptions import TokenizationError


class Tokenizer:
    def __init__(self, token_lookup: Dict) -> None:
        self.token_lookup = token_lookup

    def __lazy_tokenization(self, expression: str):
        token_list = []

        expression_simplified = simplify_expression(expression)

        # To check if single character is recognised
        digit_set = set('0123456789.')
        letter_set = set(ascii_lowercase)
        symbol_set = set(
            ''.join(filter(lambda s: not s.isalpha(), self.token_lookup.keys())))

        # To check if tokens are recognised
        recognised_tokens = set(self.token_lookup.keys()).union(set('(,)'))

        i = 0
        j = 1

        yield i, j, token_list

        while j <= len(expression_simplified):
            char = expression_simplified[i]
            if char not in '(,)':
                if char in digit_set:
                    while j < len(expression_simplified) and expression_simplified[j] in digit_set:
                        j += 1
                        yield i, j, token_list
                elif char in letter_set:
                    while j < len(expression_simplified) and expression_simplified[j] in letter_set:
                        j += 1
                        yield i, j, token_list
                elif char in symbol_set:
                    while j < len(expression_simplified) and expression_simplified[j] in symbol_set:
                        j += 1
                        yield i, j, token_list
                else:
                    raise TokenizationError(
                        expression_simplified=expression_simplified, kind='symbol', i=i, j=j)
            token = expression_simplified[i:j]
            if token in recognised_tokens or token[0] in digit_set:
                token_list.append(token)
            else:
                for k in range(j, i, -1):
                    token = expression_simplified[i:k]
                    if token in recognised_tokens:
                        token_list.append(token)
                        j = k
                        yield i, j, token_list
                        break
                else:
                    raise TokenizationError(
                        expression_simplified=expression_simplified, kind='token', i=i, j=j)
            i = j
            j += 1
            yield i, j, token_list

    def __eager_tokenization(self, expression: str) -> List[str]:
        token_list = []
        for _, _, tokens in self.__lazy_tokenization(expression=expression):
            token_list = tokens
        return token_list

    def tokenize(self, expression: str, kind: Literal['lazy', 'eager'] = 'eager'):
        if kind == 'lazy':
            return self.__lazy_tokenization(expression=expression)
        elif kind == 'eager':
            return self.__eager_tokenization(expression=expression)
        else:
            raise ValueError(f'Unknown kind \'{kind}\' received')



from typing import Optional, Union
from .node import Node


class Operand(Node):
    def __init__(self,
                 value: Union[int, float],
                 precision: int = 3,
                 symbol: Optional[str] = None,
                 **kwargs):
        super().__init__(value=value, **kwargs)
        self.precision: int = precision
        self.symbol: str = symbol if symbol is not None else self.__str__()
        self.inner_width = len(self.symbol)
        self.width = len(self.symbol)
        self.priority = 20

    def __str__(self) -> str:
        if type(self.value) is int or len(str(self.value).split('.')[1]) <= self.precision:
            return f'{self.value:^{self.width}}'
        else:
            return f'{self.value:^{self.width}.{self.precision}f}'

    def display(self) -> str:
        return ' ' * self.left_pad + self.symbol + ' ' * self.right_pad

    def __repr__(self) -> str:
        return f'Operand({self.value})'

    def isFull(self) -> bool:
        return True
